/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 *
 * The Initial Developer of the Original Code and related documentation
 * is America Online, Inc. Portions created by AOL are Copyright (C) 1999
 * America Online, Inc. All Rights Reserved.
 */

/*
 * test.c -
 *
 *      Collection of thread interface tests.  This code is somewhat sloppy
 *      but contains several examples of using conditions, mutexes,
 *      thread local storage, and creating/joining threads.
 */

#include "nsthread.h"

/*
 * Special direct include of pthread.h for compatibility tests.
 */

#ifdef _WIN32
#define PTHREAD_TEST 0
#else
#include <pthread.h>
#define PTHREAD_TEST 1
#endif

/*
 * Collection of synchronization objects for tests.
 */

static Ns_Mutex block = NULL;
static Ns_Mutex slock = NULL;
static Ns_Mutex lock  = NULL;
static Ns_Cond  cond  = NULL;
static Ns_Tls   key   = NULL;
static Ns_RWLock rwlock = NULL;
static Ns_Sema  sema  = NULL;
static Ns_Cs    cs    = NULL;
static Ns_Mutex dlock = NULL;
static Ns_Cond  dcond = NULL;
static int      dstop = 0;

/*
 * Local Prototypes
 */
static void AtExit(void);
static void MemThread(void *arg);

static void Msg(const char *fmt,...)
    NS_GNUC_PRINTF(1, 2);

static void CheckStackThread(void *UNUSED(arg)) NS_GNUC_NORETURN;
static void WorkThread(void *arg)               NS_GNUC_NORETURN;

/*
 * Msg -
 *
 *      Simple message logger with thread id and name.
 */

static void
Msg(const char *fmt,...)
{
    va_list         ap;
    time_t          now;
    char            buffer[80];

    time(&now);
    strftime(buffer, sizeof(buffer), "%d/%b/%Y:%H:%M:%S", ns_localtime(&now));
    va_start(ap, fmt);
    Ns_MutexLock(&slock);
    printf("[%s][%s]: ", Ns_ThreadGetName(), buffer);
    vfprintf(stdout, fmt, ap);
    printf("\n");
    Ns_MutexUnlock(&slock);
    va_end(ap);
}

/*
 * TlsLogArg -
 *
 *      Log and then free TLS slot data at thread exit.
 */

static void
TlsLogArg(void *arg)
{
    int            *ip = arg;

    Msg("tls cleanup %d", *ip);
    ns_free(ip);
}

/*
 * RecursiveStackCheck, CheckStackThread -
 *
 *      Thread which recursively probes stack for max depth.
 */

static uintptr_t
RecursiveStackCheck(uintptr_t n)
{
#if 0
    if (Ns_CheckStack() == NS_OK) {
        n = RecursiveStackCheck(n);
    }
#endif
    ++n;

    return n;
}

static void
CheckStackThread(void *UNUSED(arg))
{
    uintptr_t n;

    n = RecursiveStackCheck(0);
    Ns_ThreadExit((void *) n);
}

/*
 * WorkThread -
 *
 *      Thread which exercise a variety of sync objects and TLS.
 */

static void
WorkThread(void *arg)
{
    intptr_t        i = (intptr_t) arg;
    intptr_t       *ip;
    time_t          now;
    Ns_Thread       self;

    Ns_ThreadSetName("-work:%" PRIdPTR "-", i);

    if (i == 2) {
        Ns_RWLockWrLock(&rwlock);
        Msg("rwlock write acquired");
        sleep(2);
    } else {
        Ns_RWLockRdLock(&rwlock);
        Msg("rwlock read acquired");
        sleep(1);
    }
    Ns_CsEnter(&cs);
    Msg("enter critical section once");
    Ns_CsEnter(&cs);
    Msg("enter critical section twice");
    Ns_CsLeave(&cs);
    Ns_CsLeave(&cs);
    Ns_ThreadSelf(&self);
    arg = Ns_TlsGet(&key);
    Ns_SemaWait(&sema);
    Msg("got semaphore posted from main");
    if (arg == NULL) {
        arg = ns_malloc(sizeof(i));
        Ns_TlsSet(&key, arg);
    }
    ip = arg;
    *ip = i;

    if (i == 5) {
        Ns_Time         to;
        Ns_ReturnCode   st;
        char            buffer[80];

        Ns_GetTime(&to);
        Msg("time: " NS_TIME_FMT, (int64_t) to.sec, to.usec);
        Ns_IncrTime(&to, 5, 0);
        Msg("time: " NS_TIME_FMT, (int64_t) to.sec, to.usec);
        time(&now);
        strftime(buffer, sizeof(buffer), "%d/%b/%Y:%H:%M:%S", ns_localtime(&now));
        Msg("timed wait starts: %s", buffer);
        Ns_MutexLock(&lock);
        st = Ns_CondTimedWait(&cond, &lock, &to);
        Ns_MutexUnlock(&lock);
        time(&now);
        strftime(buffer, sizeof(buffer), "%d/%b/%Y:%H:%M:%S", ns_localtime(&now));
        Msg("timed wait ends: %s - status: %d", buffer, st);
    }
    if (i == 9) {
        Msg("sleep 4 seconds start");
        sleep(4);
        Msg("sleep 4 seconds done");
    }
    Ns_RWLockUnlock(&rwlock);
    Msg("rwlock unlocked");
    Msg("exiting");
    Ns_ThreadExit((void *) i);
}

/*
 * AtExit -
 *
 *      Test of atexit() handler.
 */

static void
AtExit(void)
{
    Msg("atexit handler called!");
}

/*
 * MemThread, MemTime -
 *
 *      Time allocations of malloc and MT-optmized ns_malloc
 */

#define NA 10000
#define BS (1024*16)

static int nthreads = 10;
static int memStart = 0;
static int nrunning = 0;

static void
MemThread(void *arg)
{
    int    i;
    void  *ptr;

    Ns_MutexLock(&lock);
    ++nrunning;
    Ns_CondBroadcast(&cond);
    while (memStart == 0) {
        Ns_CondWait(&cond, &lock);
    }
    Ns_MutexUnlock(&lock);

    ptr = NULL;
    for (i = 0; i < NA; ++i) {
        size_t n = (size_t)rand() % BS;
        if (arg != NULL) {
            ns_free(ptr);
            ptr = ns_malloc(n);
        } else {
            if (ptr != NULL) {
                free(ptr);
            }
            ptr = malloc(n);
        }
    }
}

static void
MemTime(int ns)
{
    Ns_Time         start, end, diff;
    int             i;
    Ns_Thread      *tids;

    tids = ns_malloc(sizeof(Ns_Thread *) * (size_t)nthreads);
    Ns_MutexLock(&lock);
    nrunning = 0;
    memStart = 0;
    Ns_MutexUnlock(&lock);
    printf("starting %d %smalloc threads...", nthreads, (ns != 0) ? "ns_" : "");
    fflush(stdout);
    for (i = 0; i < nthreads; ++i) {
        Ns_ThreadCreate(MemThread, (void *)(intptr_t) ns, 0, &tids[i]);
    }
    Ns_MutexLock(&lock);
    while (nrunning < nthreads) {
        Ns_CondWait(&cond, &lock);
    }
    printf("waiting....");
    fflush(stdout);
    memStart = 1;
    Ns_CondBroadcast(&cond);
    Ns_GetTime(&start);
    Ns_MutexUnlock(&lock);
    for (i = 0; i < nthreads; ++i) {
        Ns_ThreadJoin(&tids[i], NULL);
    }
    Ns_GetTime(&end);
    Ns_DiffTime(&end, &start, &diff);
    printf("done: " NS_TIME_FMT " sec\n", (int64_t) diff.sec, diff.usec);
}


static void
DumpString(Tcl_DString *dsPtr)
{
    char **largv;
    TCL_SIZE_T largc;

    if (Tcl_SplitList(NULL, dsPtr->string, &largc, (const char***)&largv) == TCL_OK) {
        TCL_SIZE_T i;

        for (i = 0; i < largc; ++i) {
            printf("\t%s\n", largv[i]);
        }
        Tcl_Free((char *) largv);
    }
    Tcl_DStringSetLength(dsPtr, 0);
}


static void
DumperThread(void *UNUSED(arg))
{
    Ns_Time         to;
    Tcl_DString     ds;

    Tcl_DStringInit(&ds);
    Ns_ThreadSetName("%s", "-dumper-");
    Ns_MutexLock(&block);
    Ns_MutexLock(&dlock);
    while (dstop == 0) {
        Ns_GetTime(&to);
        Ns_IncrTime(&to, 1, 0);
        (void)Ns_CondTimedWait(&dcond, &dlock, &to);
        Ns_MutexLock(&slock);
        Ns_ThreadList(&ds, NULL);
        DumpString(&ds);
        Ns_MutexList(&ds);
        DumpString(&ds);
#ifdef HAVE_TCL_GETMEMORYINFO
        Tcl_GetMemoryInfo(&ds);
#endif
        DumpString(&ds);
        Ns_MutexUnlock(&slock);
    }
    Ns_MutexUnlock(&dlock);
    Ns_MutexUnlock(&block);
}

#if PTHREAD_TEST

/*
 * Routines to test compatibility with pthread-created
 * threads, i.e., that non-Ns_ThreadCreate'd threads
 * can call Ns API's which will cleanup at thread exit.
 */

static Ns_Mutex plock = NULL;
static Ns_Cond pcond  = NULL;
static int pgo = 0;

static void
PthreadTlsCleanup(void *arg)
{
    intptr_t i = (intptr_t) arg;
    pthread_t self = pthread_self();

    printf("pthread[%" PRIxPTR "]: log: %" PRIdPTR"\n", (uintptr_t)self, i);
}

static void *
Pthread(void *arg)
{
    static Ns_Tls tls;

    /*
     * Allocate TLS first time (this is recommended TLS
     * self-initialization style.
     */

    if (tls == NULL) {
        Ns_MasterLock();
        if (tls == NULL) {
             Ns_TlsAlloc(&tls, PthreadTlsCleanup);
        }
        Ns_MasterUnlock();
    }

    Ns_TlsSet(&tls, arg);

    /*
     * Wait for exit signal from main().
     */

    Ns_MutexLock(&plock);
    while (pgo == 0) {
        Ns_CondWait(&pcond, &plock);
    }
    Ns_MutexUnlock(&plock);
    return arg;
}

#endif

/*
 * main -
 *
 *      Fire off a bunch of weird threads to exercise the thread
 *      interface.
 */

int main(int argc, char *argv[])
{
    intptr_t        i;
    Ns_Thread       threads[10];
    Ns_Thread       self, dumper;
    void *arg;
    char *p;
#if PTHREAD_TEST
    pthread_t tids[10];
#endif

    Tcl_FindExecutable(argv[0]);
    Nsthreads_LibInit();

    Ns_ThreadSetName("-main:%s-", "test");

    /*
     * Jump directly to memory test if requested.
     */

    for (i = 1; i < argc; ++i) {
        p = argv[i];
        switch (*p) {
        case 'n':
            break;
        case 'm':
            nthreads = (int)strtol(p + 1, NULL, 10);
            goto mem;
        default:
            Msg("valid arguments: ?m? ?n/NR_THREADS/?");
            return 0;
        }
    }

    Ns_ThreadCreate(DumperThread, NULL, 0, &dumper);
    Ns_MutexSetName(&lock, "startlock");
    Ns_MutexSetName(&dlock, "dumplock");
    Ns_MutexSetName(&slock, "msglock");
    Ns_MutexSetName(&block, "busylock");
    Ns_RWLockSetName2(&rwlock, "rwlock", NULL);
    Ns_CsInit(&cs);
    Ns_CondInit(&cond);
    Ns_CondInit(&dcond);
#if PTHREAD_TEST
    Ns_CondInit(&pcond);
#endif

    Ns_ThreadStackSize(81920);
    Ns_SemaInit(&sema, 3);
    Msg("sema initialized to 3");
    atexit(AtExit);
    Msg("pid = %d", getpid());
    Ns_TlsAlloc(&key, TlsLogArg);
    for (i = 0; i < 10; ++i) {
        Msg("starting work thread %" PRIdPTR, i);
        Ns_ThreadCreate(WorkThread, (void *) i, 0, &threads[i]);
    }
    sleep(1);
    /* Ns_CondSignal(&cond); */
    Ns_SemaPost(&sema, 10);
    Msg("sema post 10");
    Ns_RWLockWrLock(&rwlock);
    Msg("rwlock write locked (main thread)");
    sleep(1);
    Ns_RWLockUnlock(&rwlock);
    Msg("rwlock write unlocked (main thread)");
    for (i = 0; i < 10; ++i) {
        void *codeArg;
        Msg("waiting for thread %" PRIdPTR " to exit", i);
        Ns_ThreadJoin(&threads[i], &codeArg);
        Msg("thread %" PRIdPTR " exited - code: %" PRIuPTR, i, (uintptr_t) codeArg);
    }
#if PTHREAD_TEST
    memset(tids, 0, sizeof(tids));

    for (i = 0; i < 10; ++i) {
        pthread_create(&tids[i], NULL, Pthread, (void *) i);
        printf("pthread: create %" PRIdPTR " = %" PRIxPTR "\n", i, (uintptr_t) tids[i]);
        Ns_ThreadYield();
    }
    Ns_MutexLock(&plock);
    pgo = 1;
    Ns_MutexUnlock(&plock);
    Ns_CondBroadcast(&pcond);
    for (i = 0; i < 10; ++i) {
        pthread_join(tids[i], &arg);
        printf("pthread: join %" PRIdPTR " = %" PRIdPTR "\n", i, (intptr_t) arg);
    }
#endif
    Ns_ThreadSelf(&self);
    Ns_MutexLock(&dlock);
    dstop = 1;
    Ns_CondSignal(&dcond);
    Ns_MutexUnlock(&dlock);
    Ns_ThreadJoin(&dumper, NULL);
    Msg("threads joined");
    {
        int j;
        for (j = 0; j < 10; ++j) {
            Ns_ThreadCreate(CheckStackThread, NULL, 8192*(j+1), &threads[j]);
        }
        for (j = 0; j < 10; ++j) {
            Ns_ThreadJoin(&threads[j], &arg);
            printf("check stack %d = %" PRIdPTR "\n", j, (intptr_t) arg);
        }
    }
    /*Ns_ThreadEnum(DumpThreads, NULL);*/
    /*Ns_MutexEnum(DumpLocks, NULL);*/
mem:
    MemTime(0);
    MemTime(1);
    return 0;
}

/*
 * Local Variables:
 * mode: c
 * c-basic-offset: 4
 * fill-column: 78
 * indent-tabs-mode: nil
 * End:
 */
